/*
Copyright © 2024 Ryan D McGuire <rmcguire@libretechconsulting.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package cmd

import (
	"context"
	"os"
	"os/signal"
	"strings"
	"time"

	"github.com/rs/zerolog"
	"github.com/spf13/cobra"

	"gitea.libretechconsulting.com/50W/eia-api-go/cmd/eia-client/cmd/get"
	"gitea.libretechconsulting.com/50W/eia-api-go/cmd/eia-client/cmd/list"
	"gitea.libretechconsulting.com/50W/eia-api-go/cmd/eia-client/internal/util"
)

// rootCmd represents the base command when called without any subcommands
var rootCmd = &cobra.Command{
	Use:              "eia-client",
	Short:            "Useful utilities for EIA API v2",
	PersistentPreRun: PreRun,
}

func PreRun(cmd *cobra.Command, args []string) {
	// Set up logging
	console := zerolog.ConsoleWriter{
		Out:        os.Stdout,
		TimeFormat: time.RFC1123,
	}
	logger := zerolog.New(console).With().
		Timestamp().Logger().Level(*util.GetLogLevel(cmd))

	util.SetLogger(cmd, &logger)

	// Be quiet if we're completing
	if strings.Contains(cmd.CommandPath(), "comple") {
		cmd.Flags().Set(util.FLAG_API_LOG_LEVEL, "trace")
	} else {
		util.Logger(cmd).Debug().
			Str("logLevel", util.Logger(cmd).GetLevel().String()).
			Msg("logging configured")
	}

	util.SetClient(cmd, args)
}

// Execute adds all child commands to the root command and sets flags appropriately.
// This is called by main.main(). It only needs to happen once to the rootCmd.
func Execute() {
	ctx, cncl := signal.NotifyContext(context.Background(), os.Interrupt, os.Kill)
	defer cncl()

	err := rootCmd.ExecuteContext(ctx)
	if err != nil {
		os.Exit(1)
	}
}

func init() {
	// Enable PreRun for all commands
	cobra.EnableTraverseRunHooks = true

	// Handle command flags
	rootCmd.PersistentFlags().StringP(util.FLAG_LOG_LEVEL, "l", "",
		"Log Level, or set "+util.ENV_LOG_LEVEL+" in environment")

	rootCmd.PersistentFlags().StringP(util.FLAG_API_LOG_LEVEL, "L", "",
		"Log Level for EIA API Middleware, or set "+util.ENV_API_LOG_LEVEL+" in environment")

	rootCmd.PersistentFlags().String(util.FLAG_APIKEY,
		"", "API Key, or set "+util.ENV_API+" in environment")

	rootCmd.PersistentFlags().DurationP(util.FLAG_TMOUT, "t", util.DefRequestTimeout,
		"Request timeout, or set "+util.ENV_TMOUT+" in environment")

	// Subcommands
	rootCmd.AddCommand(list.ListCmd)
	rootCmd.AddCommand(get.GetCmd)
}
